]> git.ipfire.org Git - thirdparty/mdadm.git/blobdiff - monitor.c
Merge branch 'master' into devel-3.0
[thirdparty/mdadm.git] / monitor.c
index ffb4c9c48dc7e439cc1da352949f6c2aa6032f5b..d8ac45e8e477799b71232596b99ec2beb96faec3 100644 (file)
--- a/monitor.c
+++ b/monitor.c
@@ -47,7 +47,6 @@ static int read_attr(char *buf, int len, int fd)
        return n;
 }
 
-
 int get_resync_start(struct active_array *a)
 {
        char buf[30];
@@ -62,30 +61,6 @@ int get_resync_start(struct active_array *a)
        return 1;
 }
 
-static int attr_match(const char *attr, const char *str)
-{
-       /* See if attr, read from a sysfs file, matches
-        * str.  They must either be the same, or attr can
-        * have a trailing newline or comma
-        */
-       while (*attr && *str && *attr == *str) {
-               attr++;
-               str++;
-       }
-
-       if (*str || (*attr && *attr != ',' && *attr != '\n'))
-               return 0;
-       return 1;
-}
-
-static int match_word(const char *word, char **list)
-{
-       int n;
-       for (n=0; list[n]; n++)
-               if (attr_match(word, list[n]))
-                       break;
-       return n;
-}
 
 static enum array_state read_state(int fd)
 {
@@ -94,7 +69,7 @@ static enum array_state read_state(int fd)
 
        if (n <= 0)
                return bad_word;
-       return (enum array_state) match_word(buf, array_states);
+       return (enum array_state) sysfs_match_word(buf, array_states);
 }
 
 static enum sync_action read_action( int fd)
@@ -104,7 +79,7 @@ static enum sync_action read_action( int fd)
 
        if (n <= 0)
                return bad_action;
-       return (enum sync_action) match_word(buf, sync_actions);
+       return (enum sync_action) sysfs_match_word(buf, sync_actions);
 }
 
 int read_dev_state(int fd)
@@ -119,15 +94,15 @@ int read_dev_state(int fd)
 
        cp = buf;
        while (cp) {
-               if (attr_match(cp, "faulty"))
+               if (sysfs_attr_match(cp, "faulty"))
                        rv |= DS_FAULTY;
-               if (attr_match(cp, "in_sync"))
+               if (sysfs_attr_match(cp, "in_sync"))
                        rv |= DS_INSYNC;
-               if (attr_match(cp, "write_mostly"))
+               if (sysfs_attr_match(cp, "write_mostly"))
                        rv |= DS_WRITE_MOSTLY;
-               if (attr_match(cp, "spare"))
+               if (sysfs_attr_match(cp, "spare"))
                        rv |= DS_SPARE;
-               if (attr_match(cp, "blocked"))
+               if (sysfs_attr_match(cp, "blocked"))
                        rv |= DS_BLOCKED;
                cp = strchr(cp, ',');
                if (cp)
@@ -231,29 +206,31 @@ static int read_and_act(struct active_array *a)
                a->next_state = active;
        }
        if (a->curr_state == active_idle) {
/* Set array to 'clean' FIRST, then
* a->ss->mark_clean(a, ~0ULL);
* just ignore for now.
                 */
+               a->next_state = clean;
+       }
+       if (a->curr_state == clean) {
+               get_resync_start(a);
+               a->container->ss->set_array_state(a, 1);
        }
 
        if (a->curr_state == readonly) {
-               /* Well, I'm ready to handle things, so
-                * read-auto is OK. FIXME what if we really want
-                * readonly ???
+               /* Well, I'm ready to handle things.  If readonly
+                * wasn't requested, transition to read-auto.
                 */
-               get_resync_start(a);
-               if (a->resync_start == ~0ULL) {
-                       a->next_state = read_auto; /* array is clean */
-                       /* give the metadata a chance to force active if
-                        * we have some recovery to do.  metadata sets
-                        * resync_start to !MaxSector in this case
-                        */
-                       a->container->ss->set_array_state(a, 1);
-               }
-               if (a->resync_start != ~0ULL) {
-                       a->container->ss->set_array_state(a, 0);
-                       a->next_state = active;
+               char buf[64];
+               read_attr(buf, sizeof(buf), a->metadata_fd);
+               if (strncmp(buf, "external:-", 10) == 0) {
+                       /* explicit request for readonly array.  Leave it alone */
+                       ;
+               } else {
+                       get_resync_start(a);
+                       if (a->container->ss->set_array_state(a, 2))
+                               a->next_state = read_auto; /* array is clean */
+                       else
+                               a->next_state = active; /* Now active for recovery etc */
                }
        }
 
@@ -266,7 +243,7 @@ static int read_and_act(struct active_array *a)
                 * Just check if we need to fiddle spares.
                 */
                get_resync_start(a);
-               a->container->ss->set_array_state(a, 0);
+               a->container->ss->set_array_state(a, a->curr_state <= clean);
                check_degraded = 1;
        }
 
@@ -307,7 +284,8 @@ static int read_and_act(struct active_array *a)
        }
 
        a->container->ss->sync_metadata(a->container);
-       dprintf("%s: update[%d]: (", __func__, a->info.container_member);
+       dprintf("%s(%d): state:%s action:%s next(", __func__, a->info.container_member,
+               array_states[a->curr_state], sync_actions[a->curr_action]);
 
        /* Effect state changes in the array */
        if (a->next_state != bad_word) {
@@ -316,7 +294,7 @@ static int read_and_act(struct active_array *a)
        }
        if (a->next_action != bad_action) {
                write_attr(sync_actions[a->next_action], a->action_fd);
-               dprintf(" action:%s", array_states[a->next_state]);
+               dprintf(" action:%s", sync_actions[a->next_action]);
        }
        for (mdi = a->info.devs; mdi ; mdi = mdi->next) {
                if (mdi->next_state & DS_UNBLOCK) {
@@ -436,6 +414,7 @@ static int wait_and_act(struct supertype *container, int nowait)
        struct active_array *a, **ap;
        int rv;
        struct mdinfo *mdi;
+       static unsigned int dirty_arrays = ~0; /* start at some non-zero value */
 
        FD_ZERO(&rfds);
 
@@ -464,16 +443,20 @@ static int wait_and_act(struct supertype *container, int nowait)
                ap = &(*ap)->next;
        }
 
-       if (manager_ready && *aap == NULL) {
-               /* No interesting arrays. Lets see about exiting.
-                * Note that blocking at this point is not a problem
-                * as there are no active arrays, there is nothing that
-                * we need to be ready to do.
+       if (manager_ready && (*aap == NULL || (sigterm && !dirty_arrays))) {
+               /* No interesting arrays, or we have been told to
+                * terminate and everything is clean.  Lets see about
+                * exiting.  Note that blocking at this point is not a
+                * problem as there are no active arrays, there is
+                * nothing that we need to be ready to do.
                 */
                int fd = open(container->device_name, O_RDONLY|O_EXCL);
                if (fd >= 0 || errno != EBUSY) {
                        /* OK, we are safe to leave */
-                       dprintf("no arrays to monitor... exiting\n");
+                       if (sigterm && !dirty_arrays)
+                               dprintf("caught sigterm, all clean... exiting\n");
+                       else
+                               dprintf("no arrays to monitor... exiting\n");
                        remove_pidfile(container->devname);
                        exit_now = 1;
                        signal_manager();
@@ -508,7 +491,11 @@ static int wait_and_act(struct supertype *container, int nowait)
                container->ss->sync_metadata(container);
        }
 
+       rv = 0;
+       dirty_arrays = 0;
        for (a = *aap; a ; a = a->next) {
+               int is_dirty;
+
                if (a->replaces && !discard_this) {
                        struct active_array **ap;
                        for (ap = &a->next; *ap && *ap != a->replaces;
@@ -523,6 +510,30 @@ static int wait_and_act(struct supertype *container, int nowait)
                }
                if (a->container)
                        rv += read_and_act(a);
+               else
+                       continue;
+
+               /* when terminating stop manipulating the array after it is
+                * clean, but make sure read_and_act() is given a chance to
+                * handle 'active_idle'
+                */
+               switch (read_state(a->info.state_fd)) {
+                       case active:
+                       case active_idle:
+                       case suspended:
+                       case bad_word:
+                               is_dirty = 1;
+                               break;
+                       default:
+                               if (a->curr_state == active_idle)
+                                       is_dirty = 1;
+                               else
+                                       is_dirty = 0;
+                               break;
+               }
+               dirty_arrays += is_dirty;
+               if (sigterm && !is_dirty)
+                       a->container = NULL; /* stop touching this array */
        }
 
        /* propagate failures across container members */
@@ -231,29 +206,31 @@ static int read_and_act(struct active_array *a)
        }
        if (a->curr_state == active_idle) {
-               /* Set array to 'clean' FIRST, then
-                * a->ss->mark_clean(a, ~0ULL);
-                * just ignore for now.
+               /* Set array to 'clean' FIRST, then mark clean
+                * in the metadata
                 */